# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library module."""
from eiffellib.subscribers import RabbitMQSubscriber
from eiffellib.publishers import RabbitMQPublisher
from .lib.auth import Auth
from .lib.config import Config
from .lib.events import Events
from .lib.monitor import Monitor
from .lib.utils import Utils
from .lib.http import Http
from .lib.debug import Debug
from .lib.database import Database
from .lib.secrets import Secrets
from .lib.exceptions import (
    PublisherConfigurationMissing,
    SubscriberConfigurationMissing,
    PublisherNotStarted,
)
from .graphql.query_handler import GraphQLQueryHandler


class ETOS:  # pylint: disable=too-many-instance-attributes
    """ETOS Library."""

    publisher = None
    subscriber = None
    __auth = None
    __config = None
    __events = None
    __monitor = None
    __utils = None
    __graphql = None
    __http = None
    __debug = None
    __secrets = None
    __database = None

    def __init__(self, service_name, host, name):
        """Initialize source and service name."""
        self.config.set("source", {"name": name, "host": host})
        self.config.set("service_name", service_name)

    def __del__(self):
        """Delete references to eiffel publisher and subscriber."""
        self.config.set("publisher", None)
        self.config.set("subscriber", None)

    def start_publisher(self):
        """Start the RabbitMQ publisher using config data from ETOS library config service."""
        rabbitmq = self.config.get("rabbitmq_publisher")
        if not rabbitmq:
            raise PublisherConfigurationMissing
        self.publisher = RabbitMQPublisher(**rabbitmq)
        if not self.debug.disable_sending_events:
            self.publisher.start()
        self.config.set("publisher", self.publisher)

    def start_subscriber(self):
        """Start the RabbitMQ subscriber using config data from ETOS library config service."""
        rabbitmq = self.config.get("rabbitmq_subscriber")
        if not rabbitmq:
            raise SubscriberConfigurationMissing
        self.subscriber = RabbitMQSubscriber(**rabbitmq)
        if not self.debug.disable_receiving_events:
            self.subscriber.start()
        self.config.set("subscriber", self.subscriber)

    @property
    def debug(self):
        """Entry for debug parameters for ETOS."""
        if self.__debug is None:
            self.__debug = Debug()
        return self.__debug

    @property
    def monitor(self):
        """Entry for ETOS Library monitor service."""
        if self.__monitor is None:
            self.__monitor = Monitor()
        return self.__monitor

    @property
    def events(self):
        """Entry for ETOS Library events service. Publisher must be started."""
        if self.__events is None:
            if self.publisher is None and not self.debug.disable_sending_events:
                raise PublisherNotStarted
            self.__events = Events(self.publisher)
        return self.__events

    @property
    def auth(self):
        """Entry for ETOS Library auth service."""
        if self.__auth is None:
            self.__auth = Auth()
        return self.__auth

    @property
    def config(self):
        """Entry for ETOS Library config service."""
        if self.__config is None:
            self.__config = Config()
        return self.__config

    @property
    def utils(self):
        """Entry for ETOS Library utils service."""
        if self.__utils is None:
            self.__utils = Utils()
        return self.__utils

    @property
    def http(self):
        """Entry for ETOS Library http service."""
        if self.__http is None:
            self.__http = Http()
        return self.__http

    @property
    def graphql(self):
        """Entry for ETOS Library http service."""
        if self.__graphql is None:
            self.__graphql = GraphQLQueryHandler()
        return self.__graphql

    @property
    def database(self):
        """Entry to ETOS Library database service."""
        if self.__database is None:
            self.__database = Database()
        return self.__database

    @property
    def secrets(self):
        """Entry to ETOS Library secret service."""
        if self.__secrets is None:
            self.__secrets = Secrets()
        return self.__secrets



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS log formatter."""
import datetime
import json
import logging
import sys
import traceback


# LogRecord fields that we exclude, typically because we offer
# something better.
_EXCLUDED_FIELDS = (
    # Superfluous since we're saving the timestamp in '@timestamp'.
    "asctime",
    # The arguments to the format string can contain arbitrary objects
    # that can't be serialized to JSON. We could pass them to str(),
    # but that assumes that '%s' is used in the format string.
    "args",
    "msg",
    # Exception info is saved in exception.{type,message,stacktrace}
    # instead.
    "exc_info",
    "exc_text",
    "stack_info",
)


class EtosLogFormatter(logging.Formatter):
    """Logging formatter that produces a JSON object.

    The resulting JSON object contains the logging.LogRecord fields
    untouched. The following exception from this rule exist:

        - A '@timestamp' key with the log record's UTC timestamp in ISO8601
          format, digestible by Logstash.
        - If an exception context exists (i.e. sys.exc_info() returns
          something) the resulting JSON object will contain an 'exception'
          key pointing to an object with the following keys:

              - 'type': The fully-qualified type name (i.e. including
                 package and module).
              - 'message': The exception message.
              - 'stacktrace': The formatted stacktrace as a multiline string
                 (not ending in a newline character).
    """

    def format(self, record):
        """Serialize LogRecord data as JSON.
        Overrides the inherited behavior by ignoring any configured
        format string and storing attributes of the passed LogRecord
        object in a dictionary and serializing it to JSON. See class
        docstring for details.
        """
        fields = {
            k: v
            for k, v in record.__dict__.items()
            if k not in _EXCLUDED_FIELDS and not k.startswith("_")
        }
        fields["@timestamp"] = self.formatTime(record)
        fields["message"] = record.getMessage()
        exc_type, exc_message, stack = sys.exc_info()
        if exc_type and exc_message and stack:
            fields["exception"] = {
                "type": "{}.{}".format(exc_type.__module__, exc_type.__name__),
                "message": str(exc_message),
                # format_tb() returns a list of newline-terminated strings.
                # Don't include the final newline.
                "stacktrace": "".join(traceback.format_tb(stack)).rstrip(),
            }
        # No need to append a newline character; that's up to the handler.
        return json.dumps(fields)

    def formatTime(self, record, datefmt=None):
        """Formats the log record's timestamp in ISO8601 format, in UTC.
        Overrides the inherited behavior by always returning ISO8601
        strings without allowing custom date formats.
        Raises:
            NotImplementedError: If the `datefmt` parameter is not None.
        """
        if datefmt is not None:
            raise NotImplementedError("Only default time format is supported.")
        # Make Logstash's @timestamp parser happy by including a "T"
        # between the date and the time. Append 'Z' to make it clear
        # that the timestamp is UTC.
        return datetime.datetime.utcfromtimestamp(record.created).strftime(
            "%Y-%m-%dT%H:%M:%S.%fZ"
        )



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library authentication handler.

DEPRECATED: Use kubernetes secrets instead!
"""
import configparser
from cryptography.fernet import Fernet


class Auth:
    """Authentication helper for ETOS.

    Example:
        auth = Auth()
        auth.load_password_file("password.secret")
        username, password = auth.credentials("my_service")
    """

    config = None
    key = None

    def load_password_file(self, _file):
        """Load a password file into memory.

        :param _file: Path to file to load.
        :type _file: str
        """
        self.config = configparser.ConfigParser()
        self.config.read(_file)
        self.key = self.config["global"]["key"]

    def decrypt(self, token):
        """Decrypt a password hash into a password.

        :param token: Password hash to decrypt.
        :type token: str
        :return: Password in clear-text.
        :rtype: str
        """
        cipher_suite = Fernet(self.key)
        uncipher_text = cipher_suite.decrypt(token.encode("utf-8"))
        return bytes(uncipher_text).decode("utf-8")

    def credentials(self, service):
        """Get credentials from a specific service defined in password file.

        :param service: Which service to get credentials for.
        :type service: str
        :return: Username and password for service.
        :rtype: tuple
        """
        service = self.config[service]
        return service.get("username"), self.decrypt(service.get("password"))



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library GraphQL query handler module."""
from gql import Client, gql
from gql.transport.requests import RequestsHTTPTransport
from etos_lib.lib.utils import Utils
from etos_lib.lib.debug import Debug


class GraphQLQueryHandler:
    """Create and send GraphQL queries."""

    __client = None
    __query = None

    def __init__(self):
        """Graphql query handler."""
        self.transport_protocol = RequestsHTTPTransport
        self.utils = Utils()
        self.debug = Debug()

    @property
    def transport(self):
        """Transport protocol used for graphql."""
        return self.transport_protocol(self.debug.graphql_server)

    @property
    def client(self):
        """Graphql client."""
        if self.__client is None:
            self.__client = Client(transport=self.transport)
        return self.__client

    @staticmethod
    def query(query):
        """Create a query for graphql.

        :param query: Query to make.
        :type query: str
        :return: Graphql formatted query.
        :rtype: :obj:`gql`
        """
        return gql(query)

    def execute(self, query):
        """Execute query against graphql.

        :param query: Query to make.
        :type query: str
        :return: Result from query.
        :rtype: dict
        """
        return self.client.execute(self.query(query))

    def search_for_nodes(self, response, *nodes):
        """Search for nodes in a GraphQL response. Iterator.

        :param response: GraphQL response dictionary.
        :type response: dict
        :param nodes: Nodes to search for.
                      They are represented as keys in the response dict.
        :type nodes: list
        :return: Node name and node dictionary.
        :rtype: tuple
        """
        for node_name, edge in self.utils.search(response, *nodes):
            if edge.get("edges") is not None:
                for node in edge.get("edges"):
                    yield node_name, node["node"]
            else:
                yield node_name, edge



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library exceptions."""


class ETOSException(Exception):
    """Base ETOS library exception."""


class SubscriberConfigurationMissing(ETOSException):
    """Exception for when SubscriberConfiguration is missing."""

    def __init__(self):
        """Initialize with base exception message."""
        super().__init__("Missing rabbitmq_subscriber in ETOS config.")


class PublisherConfigurationMissing(ETOSException):
    """Exception for when PublisherConfiguration is missing."""

    def __init__(self):
        """Initialize with base exception message."""
        super().__init__("Missing rabbitmq_publisher in ETOS config.")


class PublisherNotStarted(ETOSException):
    """Exception for when Publisher has not started yet."""

    def __init__(self):
        """Initialize with base exception message."""
        super().__init__("RabbitMQ publisher is not started.")


class SubscriberNotStarted(ETOSException):
    """Exception for when Subscriber has not started yet."""

    def __init__(self):
        """Initialize with base exception message."""
        super().__init__("RabbitMQ subscriber is not started.")


class PublisherDied(ETOSException):
    """Published has died."""

    def __init__(self):
        """Initialize with base exception message."""
        super().__init__("RabbitMQ publisher has died.")


class SubscriberDied(ETOSException):
    """Subscriber has died."""

    def __init__(self):
        """Initialize with base exception message."""
        super().__init__("RabbitMQ subscriber has died.")


class AliveMessageMissing(ETOSException):
    """Exception for when Alive mssage is missing."""

    def __init__(self):
        """Initialize with base exception message."""
        super().__init__("Alive message cannot be None if 'with_announcement' is True")



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library database handler."""
import os
from redis.sentinel import Sentinel
from .debug import Debug


class Database:
    """ETOS database handler."""

    connected = False
    __database = None
    __reader = None
    __writer = None

    def __init__(self, expire=3600):
        """Load password and host.

        :params expire: How long, in seconds, to persist data keys.
                        Set to None to disable expiration.
        :type expire: int
        """
        self.password = os.getenv("ETOS_DATABASE_PASSWORD")
        self.host = [(Debug().database_host, Debug().database_port)]
        self.expire = expire

    def write(self, key, value):
        """Write a single key to database.

        :param key: Key to write to.
        :type key: str
        :param value: Value to write.
        :type value: any
        :return: Response from redis.
        :rtype: :obj:`redis.Response`
        """
        result = self.writer.set(key, value)
        if self.expire is not None:
            self.writer.expire(key, self.expire)
        return result

    def append(self, key, value):
        """Add a value to key.

        :param key: Key to write to.
        :type key: str
        :param value: Value to write.
        :type value: any
        :return: Response from redis.
        :rtype: :obj:`redis.Response`
        """
        result = self.writer.zadd(key, value)
        if self.expire is not None:
            self.writer.expire(key, self.expire)
        return result

    def read_list(self, key, list_range=(0, -1)):
        """Read a key as list within range.

        :param key: Key to read from.
        :type key: str
        :param list_range: Range of list to return. Defaults to all values.
        :type list_range: tuple
        :return: Value of key.
        :rytpe: any
        """
        return self.reader.zrange(key, *list_range)

    def read(self, key):
        """Read a single key from database.

        :param key: Key to read from.
        :type key: str
        :return: Value of key.
        :rytpe: any
        """
        return self.reader.get(key)

    def remove(self, key):
        """Remove a key from database.

        :param key: Key to remove.
        :type key: str
        :return: Response from redis.
        :rtype: :obj:`redis.Response`
        """
        return self.writer.delete(key)

    @property
    def writer(self):
        """Database writer object."""
        if self.__writer is None:
            self._connect()
            self.__writer = self.__database.master_for("mymaster")
        return self.__writer

    @property
    def reader(self):
        """Database reader object."""
        if self.__reader is None:
            self._connect()
            self.__reader = self.__database.slave_for("mymaster")
        return self.__reader

    def _connect(self):
        """Connect to database."""
        if not self.connected:
            if self.password is None:
                self.__database = Sentinel(self.host)
            else:
                self.__database = Sentinel(
                    self.host,
                    sentinel_kwargs={"password": self.password},
                    password=self.password,
                )



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library event helper module."""
from eiffellib.events import EiffelActivityTriggeredEvent
from eiffellib.events import EiffelActivityStartedEvent
from eiffellib.events import EiffelActivityFinishedEvent
from eiffellib.events import EiffelActivityCanceledEvent
from eiffellib.events import EiffelAnnouncementPublishedEvent
from eiffellib.events import EiffelConfidenceLevelModifiedEvent
from eiffellib.events import EiffelEnvironmentDefinedEvent
from eiffellib.events import EiffelTestSuiteStartedEvent
from eiffellib.events import EiffelTestSuiteFinishedEvent
from eiffellib.events import EiffelTestExecutionRecipeCollectionCreatedEvent
from eiffellib.events import EiffelTestCaseTriggeredEvent
from eiffellib.events import EiffelTestCaseStartedEvent
from eiffellib.events import EiffelTestCaseFinishedEvent
from eiffellib.events import EiffelArtifactCreatedEvent
from eiffellib.events import EiffelArtifactPublishedEvent
from eiffellib.events import EiffelCompositionDefinedEvent
from .debug import Debug


class Events:
    """Helper class for sending eiffel events."""

    def __init__(self, publisher):
        """Initialize event helper."""
        self.publisher = publisher
        self.debug = Debug()

    def __del__(self):
        """Delete reference to eiffel publisher."""
        self.publisher = None

    def send(self, event, links, data):
        """Build an event and send it with an eiffel publisher.

        :param event: Initialized event to send.
        :type event: :obj:`eiffel.events.base_event.BaseEvent`
        :param links: Dictionary of links to add to event.
        :type links: dict
        :param data: Dictionary of data to add to event.
        :type data: dict
        :return: The event that was created with data and links added.
        :rtype: :obj:`eiffel.events.base_event.BaseEvent`
        """
        for key, value in links.items():
            if isinstance(value, list):
                for link in value:
                    event.links.add(key.upper(), link)
            else:
                event.links.add(key.upper(), value)
        for key, value in data.items():
            event.data.add(key, value)
        event.validate()
        self.debug.events_published.append(event)
        if not self.debug.disable_sending_events:
            self.publisher.send_event(event)
        return event

    def send_activity_triggered(self, name, links=None, **optional):
        """Send activity triggered event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelActivityTriggeredEvent.md

        :param name: Name of the activity
        :type name: str
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        :return: The event that was created with data and links added.
        :rtype: :obj:`eiffel.events.EiffelActivityTriggeredEvent`
        """
        links = links if links is not None else {}
        data = {"name": name}
        data.update(**optional)
        return self.send(EiffelActivityTriggeredEvent(), links, data)

    def send_activity_canceled(self, triggered, links=None, **optional):
        """Send activity canceled event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelActivityCanceledEvent.md

        :param triggered: Event ID of activity triggered event which is canceled.
        :type triggered: str
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        links.update({"ACTIVITY_EXECUTION": triggered})
        data = optional
        return self.send(EiffelActivityCanceledEvent(), links, data)

    def send_activity_started(self, triggered, links=None, **optional):
        """Send activity started event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelActivityStartedEvent.md

        :param triggered: Event ID of activity triggered event which is started.
        :type triggered: str
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        links.update({"ACTIVITY_EXECUTION": triggered})
        data = optional
        return self.send(EiffelActivityStartedEvent(), links, data)

    def send_activity_finished(self, triggered, outcome, links=None, **optional):
        """Send activity finished event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelActivityFinishedEvent.md

        :param triggered: Event ID of activity triggered event which is finished.
        :type triggered: str
        :param outcome: Outcome of the activity.
        :type outcome: dict
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        links.update({"ACTIVITY_EXECUTION": triggered})
        data = {"outcome": outcome}
        data.update(**optional)
        return self.send(EiffelActivityFinishedEvent(), links, data)

    def send_environment_defined(self, name, links=None, **optional):
        """Send environment defined event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelEnvironmentDefinedEvent.md

        :param name: Name of environment.
        :type name: str
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        if (
            optional.get("image") is None
            and optional.get("host") is None
            and optional.get("uri") is None
        ):
            raise Exception("At least one of 'host', 'image' or 'uri' must be provided")
        links = links if links is not None else {}
        data = {"name": name}
        data.update(**optional)
        return self.send(EiffelEnvironmentDefinedEvent(), links, data)

    def send_test_suite_started(self, name, links=None, **optional):
        """Publish a test suite started event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelTestSuiteStartedEvent.md

        :param name: Name of testsuite.
        :type name: str
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        data = {"name": name}
        data.update(**optional)
        return self.send(EiffelTestSuiteStartedEvent(), links, data)

    def send_test_suite_finished(self, test_suite, links=None, **optional):
        """Publish a test suite finished event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelTestSuiteFinishedEvent.md

        :param test_suite: A reference to the started test suite.
        :type test_suite: :obj:`eiffel.events.base_event.BaseEvent`
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        links.update({"TEST_SUITE_EXECUTION": test_suite})
        data = optional
        return self.send(EiffelTestSuiteFinishedEvent(), links, data)

    def send_announcement_published(
        self, heading, body, severity, links=None, **optional
    ):
        """Publish an announcement event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelAnnouncementPublishedEvent.md

        :param heading: Heading for the announcement.
        :type heading: str
        :param body: Body for the announcement.
        :type body: str
        :param severity: Severity of the incident.
        :type severity: str
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        data = {"heading": heading, "body": body, "severity": severity}
        data.update(**optional)
        return self.send(EiffelAnnouncementPublishedEvent(), links, data)

    def send_test_execution_recipe_collection_created(
        self, selection_strategy, links=None, **optional
    ):
        """Publish a TERCC event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelTestExecutionRecipeCollectionCreatedEvent.md

        :param selection_strategy: Selection strategy used by tercc
        :type selection_strategy: dict
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        if optional.get("batches") is None and optional.get("batchesUri") is None:
            raise Exception(
                "At least one of 'batches' or 'batchesUri' must be provided"
            )
        links = links if links is not None else {}
        data = {"selectionStrategy": selection_strategy}
        data.update(**optional)
        return self.send(EiffelTestExecutionRecipeCollectionCreatedEvent(), links, data)

    def send_confidence_level_modified(self, name, value, links=None, **optional):
        """Publish a confidence level event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelConfidenceLevelModifiedEvent.md

        :param name: Name of confidence level.
        :type name: str
        :param value: Value of confidence.
        :type value: str
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        data = {"name": name, "value": value}
        data.update(**optional)
        return self.send(EiffelConfidenceLevelModifiedEvent(), links, data)

    def send_test_case_triggered(self, test_case, iut, links=None, **optional):
        """Publish a confidence level event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelTestCaseTriggeredEvent.md

        :param test_case: TestCase that has been triggered.
        :type test_case: dict
        :param iut: Item under test.
        :type iut: :obj:`eiffel.events.base_event.BaseEvent`
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        links.update({"IUT": iut})
        data = {"testCase": test_case}
        data.update(**optional)
        return self.send(EiffelTestCaseTriggeredEvent(), links, data)

    def send_test_case_started(self, test_case, links=None, **optional):
        """Publish a confidence level event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelTestCaseStartedEvent.md

        :param test_case: Item under test.
        :type test_case: :obj:`eiffel.events.base_event.BaseEvent`
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        links.update({"TEST_CASE_EXECUTION": test_case})
        data = optional
        return self.send(EiffelTestCaseStartedEvent(), links, data)

    def send_test_case_finished(self, test_case, outcome, links=None, **optional):
        """Publish a confidence level event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelTestCaseFinishedEvent.md

        :param test_case: Item under test.
        :type test_case: :obj:`eiffel.events.base_event.BaseEvent`
        :param outcome: Outcome of the test case.
        :type outcome: dict
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        links.update({"TEST_CASE_EXECUTION": test_case})
        data = {"outcome": outcome}
        data.update(**optional)
        return self.send(EiffelTestCaseFinishedEvent(), links, data)

    def send_artifact_created_event(self, identity, links=None, **optional):
        """Publish an artifact created event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelArtifactCreatedEvent.md

        :param identity: PURL identity specification
        :type identity: str
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        data = {"identity": identity}
        data.update(**optional)
        return self.send(EiffelArtifactCreatedEvent(), links, data)

    def send_artifact_published_event(
        self, locations, artifact, links=None, **optional
    ):
        """Publish an artifact created event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelArtifactPublishedEvent.md

        :param locations: Locations for this artifact.
        :type locations: list
        :param artifact: Artifact created link.
        :type artifact: :obj:`eiffel.events.base_event.BaseEvent`
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        links.update({"ARTIFACT": artifact})
        data = {"locations": locations}
        data.update(**optional)
        return self.send(EiffelArtifactPublishedEvent(), links, data)

    def send_composition_defined_event(self, name, links=None, **optional):
        """Publish a composition defined event.

        https://github.com/eiffel-community/eiffel/blob/master/eiffel-vocabulary/EiffelCompositionDefinedEvent.md

        :param name: Name of composition
        :type name: str
        :param links: Optional links to add to event.
        :type links: dict
        :param optional: Dictionary of optional data to add.
        :type optional: dict
        """
        links = links if links is not None else {}
        data = {"name": name}
        data.update(**optional)
        return self.send(EiffelCompositionDefinedEvent(), links, data)



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library base kubernetes library."""
import os
from uuid import uuid4
import yaml
from kubernetes import client, config  # pylint:disable=no-name-in-module


class Kubernetes:
    """Base kubernetes library."""

    __batch = None
    __core = None
    __apps = None
    __extensions = None

    def __init__(
        self, namespace=os.getenv("ETOS_NAMESPACE"), context=None, in_cluster=True
    ):
        """Initialize kubernetes library and load kubernetes configuration.

        :param namespace: Which namespace to operate in.
        :type namespace: str
        :param context: From which Kubernetes context should we load config from.
                        This is only relevant when running locally (in_cluster=False).
        :type context: str
        :param in_cluster: Whether or not the kubernetes service is executing in a cluster.
                           Set to False when running locally, set to True when running in
                           kubernetes.
        :type in_cluster: bool
        """
        self.namespace = namespace
        self.context = context
        self.in_cluster = in_cluster
        self._load_config()

    def _load_config(self):
        """Load kubernetes configuration."""
        if self.in_cluster:
            config.load_incluster_config()
        else:
            config.load_kube_config(context=self.context)

    @staticmethod
    def load_yaml(data):
        """Load a YAML file.

        :param data: YAML data to load.
        :type data: str
        :return: Loaded yaml data.
        :rtype: :obj:`yaml`
        """
        return yaml.load(data, Loader=yaml.Loader)

    @staticmethod
    def uniqueify(name):
        """Make a name unique.

        :param name: Name to make unique.
        :type name: str
        :return: New, unique, name.
        :rtype: str
        """
        return "{}-{}".format(name, uuid4().hex[:10].upper())

    @property
    def apps_v1(self):
        """AppsV1Api for Kubernetes."""
        if self.__apps is None:
            self.__apps = client.AppsV1Api()
        return self.__apps

    @property
    def batch_v1(self):
        """BatchV1Api for Kubernetes."""
        if self.__batch is None:
            self.__batch = client.BatchV1Api()
        return self.__batch

    @property
    def extensions(self):
        """Create a ExtensionsV1beta1Api endpoint."""
        if self.__extensions is None:
            self.__extensions = client.ExtensionsV1beta1Api()
        return self.__extensions

    @property
    def core(self):
        """Create a CoreV1Api endpoint."""
        if self.__core is None:
            self.__core = client.CoreV1Api()
        return self.__core



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library config."""
from pprint import pprint
import os


class Config:
    """ETOS configuration storage."""

    config = {
        "proxies": {
            "http": os.getenv("HTTP_PROXY_OVERRIDE"),
            "https": os.getenv("HTTPS_PROXY_OVERRIDE"),
        },
        "rabbitmq": {},
        "rabbitmq_publisher": {},
        "rabbitmq_subscriber": {},
    }

    def reset(self):
        """Reset configuration back to default."""
        self.config.clear()
        self.config["proxies"] = {
            "http": os.getenv("HTTP_PROXY_OVERRIDE"),
            "https": os.getenv("HTTPS_PROXY_OVERRIDE"),
        }
        self.config["rabbitmq"] = {}
        self.config["rabbitmq_publisher"] = {}
        self.config["rabbitmq_subscriber"] = {}

    def get(self, key):
        """Get from Config."""
        return self.config.get(key)

    def get_config_or_env(self, key):
        """Get from Config or environment. Config is prioritized.

        Note that if taken from environment, it will be stored in config
        for future use.
        """
        value = self.config.get(key)
        if value is None:
            value = os.getenv(key)
            if value is not None:
                self.config[key] = value
        return self.config.get(key)

    def set(self, key, value):
        """Set a key in configuration."""
        self.config[key] = value

    @classmethod
    def dump(cls):
        """Dump the values of the configuration."""
        pprint(cls.config)

    @property
    def rabbitmq_publisher(self):
        """Rabbitmq Publisher data."""
        return self.config.get("rabbitmq_publisher")

    @property
    def rabbitmq_subscriber(self):
        """Rabbitmq Subscriber data."""
        return self.config.get("rabbitmq_subscriber")

    def rabbitmq_from_environment(self):
        """Load RabbitMQ data from environment variables."""
        ssl = os.getenv("RABBITMQ_SSL", "true") == "true"
        data = {
            "host": os.getenv("RABBITMQ_HOST", "127.0.0.1"),
            "exchange": os.getenv("RABBITMQ_EXCHANGE", "eiffel"),
            "username": os.getenv("RABBITMQ_USERNAME", None),
            "password": os.getenv("RABBITMQ_PASSWORD", None),
            "port": int(os.getenv("RABBITMQ_PORT", "5672")),
            "vhost": os.getenv("RABBITMQ_VHOST", None),
            "ssl": ssl,
        }
        self.set("rabbitmq", data)

    def rabbitmq_subscriber_from_environment(self):
        """Load RabbitMQ subscriber data from environment variables and set in config."""
        if not self.get("rabbitmq"):
            self.rabbitmq_from_environment()
        data = {"queue": os.getenv("RABBITMQ_QUEUE", None), "routing_key": "#"}
        data.update(**self.get("rabbitmq").copy())
        self.set("rabbitmq_subscriber", data)

    def rabbitmq_publisher_from_environment(self):
        """Load RabbitMQ publisher data from environment variables and set in config.

        Eiffel source is loaded from configuration file.
        """
        if not self.get("rabbitmq"):
            self.rabbitmq_from_environment()
        data = {
            "source": self.get("source"),
        }
        data.update(**self.get("rabbitmq").copy())
        self.set("rabbitmq_publisher", data)



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library utilities."""
import os
import sys
import time
import signal
import logging
import subprocess
import traceback
from contextlib import contextmanager
from .debug import Debug

# pylint:disable=redefined-outer-name


class SubprocessReadTimeout(Exception):
    """Timeout on reading from subprocess."""


class Utils:
    """ETOS utility methods."""

    logger = logging.getLogger("Utils")

    def __init__(self):
        """Initialize debug instance."""
        self.debug = Debug()

    def wait(self, method, timeout=None, interval=5, **kwargs):
        """Iterate over result from method call.

        :param method: Method to call.
        :type method: :meth:
        :param timeout: How long, in seconds, to iterate.
        :type timeout: int or None
        :param interval: How long, in seconds, to wait between method calls.
        :type interval: int
        :param kwargs: Keyword arguments to pass to method call.
        :type kwargs: dict
        """
        if timeout is None:
            timeout = self.debug.default_wait_timeout
        end = time.time() + timeout
        while time.time() < end:
            try:
                yield method(**kwargs)
            except Exception as exception:  # pylint:disable=broad-except
                if isinstance(exception, GeneratorExit):
                    break
                traceback.print_exc()
                self.logger.info("Retrying..")
            time.sleep(interval)

    def search(self, dictionary, *keys):
        """Recursively search for keys in a nested dictionary.

        :param dictionary: Dictionary to search in.
        :type dictionary: dict
        :param keys: Keys to search for.
        :type keys: list
        :return: key and value matching the search.
        :rtype: tuple
        """
        for key, value in dictionary.items():
            if key in keys:
                yield key, value
            if isinstance(value, dict):
                for key, value in self.search(value, *keys):
                    yield key, value
            elif isinstance(value, (list, tuple)):
                for item in value:
                    if isinstance(item, dict):
                        for key, value in self.search(item, *keys):
                            yield key, value

    def search_and_replace(self, dictionary, replace_with, *keys):
        """Recursively search for keys in a nested dictionary and replace them.

        :param dictionary: Dictionary to search in.
        :type dictionary: dict
        :param replace_with: Value to replace key with.
        :type replace_with: *
        :param keys: Keys to search for.
        :type keys: list
        """
        for key, value in dictionary.items():
            if key in keys:
                dictionary[key] = replace_with
            if isinstance(value, dict):
                self.search_and_replace(value, replace_with, *keys)
            elif isinstance(value, (list, tuple)):
                for item in value:
                    if isinstance(item, dict):
                        self.search_and_replace(item, replace_with, *keys)

    @contextmanager
    def chdir(self, path):
        """Change directory to path. Create path if it does not exist.

        Moves back to the previous path after exiting the context.

        Example:
            >>> with chdir("/home/test"):
            >>>     print(os.getcwd())
            /home/test

        :param path: Which path to change dir to.
        :type path: str
        """
        old_path = os.getcwd()

        self.logger.debug("Moving to directory: %s", path)

        if not os.path.isdir(path):
            self.logger.debug("Directory does not exist. Creating.")
            os.makedirs(path)

        os.chdir(path)
        yield

        self.logger.debug("Returning to: %s", old_path)
        os.chdir(old_path)

    @contextmanager
    def no_proxy(self, env=None):
        """Remove proxy from environment and reset it after.

        Example:
            >>> with no_proxy():
            >>>     print(os.environ.get("http_proxy"))
            >>>     print(os.environ.get("https_proxy"))
            >>>     print(os.environ.get("HTTP_PROXY"))
            >>>     print(os.environ.get("HTTPS_PROXY"))
            >>> print(os.environ.get("http_proxy"))
            None
            None
            None
            None
            http://proxy:1234

        Example:
            >>> my_env = os.environ.copy()
            >>> with no_proxy(my_env):
            >>>     print(os.environ.get("http_proxy"))
            >>>     print(my_env.get("http_proxy"))
            http://proxy:1234
            None

        :param env: Environment dict to remove proxy from. Optional.
        :type env: dict or None
        """
        if env is None:
            env = os.environ
        self.logger.debug("Disabling proxy")
        old_proxies = {
            "https_proxy": env.get("https_proxy"),
            "http_proxy": env.get("http_proxy"),
            "HTTPS_PROXY": env.get("HTTPS_PROXY"),
            "HTTP_PROXY": env.get("HTTP_PROXY"),
        }
        self.logger.debug("Old proxies: %s", old_proxies)
        for key, _ in old_proxies.items():
            try:
                env.pop(key)
            except KeyError:
                pass
        yield
        self.logger.debug("Restoring proxy")
        for key, value in old_proxies.items():
            if value is not None:
                env[key] = value

    @staticmethod
    def eiffel_link(event, link_type):
        """Get a LINK from an event based on type.

        Example:
            eiffel_link(tercc, "CAUSE")
        Note:
            Only returns the very first link.
            If there are multiple links you might not get the one you expect.

        :param event: Event to search in.
        :type event: :obj:`eiffel.events.BaseEiffelEvent`
        :param link_type: Type of link to search.
        :type link_type: str
        :return: The target in the link found.
        :rtype: str
        """
        for link in event.links.links:
            if link.get("type") == link_type:
                return link.get("target")
        return None

    @staticmethod
    def subprocess_signal_handler(signum, frame):  # pylint:disable=unused-argument
        """Raise subprocess read timeout."""
        raise SubprocessReadTimeout("Timeout while reading subprocess stdout.")

    def call(
        self, cmd, shell=False, env=None, executable=None, output=None, wait_output=True
    ):  # pylint:disable=too-many-arguments
        """Call a system command.

        :param cmd: Command to run.
        :type cmd: list
        :param env: Override subprocess environment.
        :type env: dict
        :param executable: Override subprocess executable.
        :type executable: str
        :param output: Path to a file to write stdout to.
        :type output: str
        :param wait_output: Whether or not to wait for output.
                            Some commands can fail in a non-interactive
                            shell due to waiting for 'readline' forever.
                            Set this to False on commands that we're
                            not in control of.
        :type wait_output: boolean
        :return: Result and output from command.
        :rtype: tuple
        """
        out = []
        for _, line in self.iterable_call(
            cmd, shell, env, executable, output, wait_output
        ):
            if isinstance(line, str):
                out.append(line)
            else:
                success = line
                break
        return success, out

    def iterable_call(
        self, cmd, shell=False, env=None, executable=None, output=None, wait_output=True
    ):  # pylint:disable=too-many-arguments
        """Call a system command and yield the output.

        :param cmd: Command to run.
        :type cmd: list
        :param env: Override subprocess environment.
        :type env: dict
        :param executable: Override subprocess executable.
        :type executable: str
        :param output: Path to a file to write stdout to.
        :type output: str
        :param wait_output: Whether or not to wait for output.
                            Some commands can fail in a non-interactive
                            shell due to waiting for 'readline' forever.
                            Set this to False on commands that we're
                            not in control of.
        :type wait_output: boolean
        :return: Result and output from command.
        :rtype: tuple
        """
        self.logger.debug("Running command: %s", " ".join(cmd))
        if shell:
            cmd = " ".join(cmd)
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=shell,
            env=env,
            executable=executable,
        )

        signal.signal(signal.SIGALRM, self.subprocess_signal_handler)
        output_file = None
        try:
            if output:
                output_file = open(output, "w")
            # Make sure you can read all output with 'docker logs'
            for line in iter(proc.stdout.readline, b""):
                yield proc, line.decode("utf-8").strip()
                sys.stdout.write(line.decode("utf-8"))
                signal.alarm(0)

                sys.stdout.flush()
                if output_file:
                    output_file.write(line.decode("utf-8"))

                if not wait_output:
                    signal.alarm(5)
        except SubprocessReadTimeout:
            pass
        finally:
            if output_file:
                output_file.close()

        _, err = proc.communicate()
        if err is not None:
            self.logger.debug(err.decode("utf-8"))
        self.logger.debug("Return code: %s (0=Good >0=Bad)", proc.returncode)

        # Unix return code 0 = success >0 = failure.
        # Python int 0 = failure >0 = success.
        # Converting unix return code to python bool.
        success = not proc.returncode

        yield proc, success



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library kubernets jobs handler module."""
import time
from .base import Kubernetes


class Job(Kubernetes):
    """Creating and managing jobs in kubernetes."""

    def create_job(self, body):
        """Create a namespaced kubernetes job.

        :param body: Yaml data to create job from.
        :type body: :obj:`yaml`
        :return: Response from kubernetes.
        :rtype: :obj:
        """
        return self.batch_v1.create_namespaced_job(self.namespace, body=body)

    def wait_for_job_started(self, job_name, timeout=300):
        """Wait for a kubernetes job status to become 'active'.

        :param job_name: Name of job to wait for.
        :type job_name: str
        :param timeout: How long, in seconds, to wait for job to start.
                        Default: 5 minutes.
        :type timeout: int
        :return: Job response.
        :rtype: dict
        """
        timeout = time.time() + timeout
        while time.time() < timeout:
            response = self.batch_v1.read_namespaced_job_status(
                job_name, self.namespace
            )
            status = response.status
            if status.active is not None:
                print("Started at: {}".format(str(status.start_time)))
                break
            time.sleep(1)
        return response

    def wait_for_job_finished(self, job_name, timeout=(3600 * 10)):
        """Wait for a kubernetes job status to become 'failed' or 'succeeded'.

        :param job_name: Name of job to wait for.
        :type job_name: str
        :param timeout: How long, in seconds, to wait for job to start.
                        Default: 10 hours.
        :type timeout: int
        :return: Result of the job.
        :rtype: bool
        """
        timeout = time.time() + timeout
        result = False
        while time.time() < timeout:
            response = self.batch_v1.read_namespaced_job_status(
                job_name, self.namespace
            )
            status = response.status
            # pylint:disable=no-else-break
            if status.failed is not None:
                print("Finished, but failed: {}".format(status))
                result = False
                break
            elif status.succeeded is not None:
                print("Finished successfully")
                result = True
                break
            time.sleep(1)
        return result

    def delete_job(self, job_name):
        """Delete a kubernetes job.

        :param job_name: Name of job to remove.
        :type job_name: str
        """
        self.batch_v1.delete_namespaced_job(job_name, self.namespace, {})



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library monitoring system."""
import time
import sys
import signal
from contextlib import contextmanager
from .config import Config
from .events import Events
from .debug import Debug
from .exceptions import (
    AliveMessageMissing,
    PublisherNotStarted,
    SubscriberDied,
    PublisherDied,
)


class Monitor:
    """Monitor processes, health check and send status messages."""

    announcement = None

    def __init__(self):
        """Initialize debug and config instances."""
        self.debug = Debug()
        self.config = Config()

    def __del__(self):
        """Delete config reference."""
        self.config = None

    def keep_alive(self, alive_message=None):
        """Keep subscriber and publishers alive forever.

        :param alive_message: Body of the announcement published event.
        :type alive_message: str
        """
        signal.signal(signal.SIGINT, self._handle_shutdown)
        signal.signal(signal.SIGTERM, self._handle_shutdown)

        if not self.debug.disable_monitor_announce and alive_message is None:
            raise AliveMessageMissing
        with self.announce(alive_message):
            while True:
                time.sleep(0.1)
                try:
                    self.check_publisher_liveness()
                except PublisherDied:
                    self.probe()
                    raise
                try:
                    self.check_subscriber_liveness()
                except SubscriberDied:
                    self.probe()
                    raise

    def probe(self):
        """Probe the environment for debug data."""
        self.config.dump()

    @contextmanager
    def announce(self, body):
        """Announce the status of this service if 'with_announcement' is True.

        :param body: Body of the announcement event.
        :type body: str
        """
        if self.debug.disable_monitor_announce:
            yield

        publisher = self.config.get("publisher")
        if publisher is None:
            raise PublisherNotStarted

        events = Events(publisher)
        heading = "{} is up and running.".format(self.config.get("service_name"))
        self.announcement = events.send_announcement_published(heading, body, "MINOR")
        try:
            yield
        except Exception as exception:  # pylint:disable=broad-except
            heading = "{} stopped running unexpectedly.".format(
                self.config.get("service_name")
            )
            body = ("{}\n").format(str(exception))
            events.send_announcement_published(
                heading,
                body,
                "CRITICAL",
                links={"modified_announcement": self.announcement},
            )

    def _handle_shutdown(self, *_, **__):
        """Signal handler for SIGINT and SIGTERM.

        Send an announcement that the service has shut down, if 'with_announcement' is set.
        sys.exit(0) will exit this service cleanly, this will make sure that kubernetes does not
        restart the service.
        """
        if not self.debug.disable_monitor_announce:
            publisher = self.config.get("publisher")
            if publisher is None:
                raise PublisherNotStarted

            heading = "{} was shut down.".format(self.config.get("service_name"))
            body = ("{} has stopped sending and listening to Eiffel events.").format(
                self.config.get("service_name")
            )
            events = Events(publisher)
            events.send_announcement_published(
                heading,
                body,
                "CLOSED",
                links={"modified_announcement": self.announcement},
            )
        sys.exit(0)

    def check_publisher_liveness(self):
        """Check if publisher is alive and well."""

    def check_subscriber_liveness(self):
        """Check if subscriber is alive and well."""
        subscriber = self.config.get("subscriber")
        if subscriber is not None:
            if not subscriber.is_alive():
                raise SubscriberDied



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS filter."""
import logging


class EtosFilter(logging.Filter):  # pylint:disable=too-few-public-methods
    """Filter for adding extra application specific data to log messages."""

    def __init__(self, application, version, environment, local):
        """Initialize with a few ETOS application fields.

        :param application: Name of application.
        :type application: str
        :param version: Version of application.
        :type version: str
        :param environment: In which environment is this executing.
        :type environment: str
        :param local: Thread-local configuration information.
        :type local: :obj:`threading.local`
        """
        self.application = application
        self.version = version
        self.environment = environment
        self.local = local
        super().__init__()

    def filter(self, record):
        """Add contextual data to log record.

        :param record: Log record to add data to.
        :type record: :obj:`logging.LogRecord`
        :return: True
        :rtype: bool
        """
        record.application = self.application
        record.version = self.version
        record.environment = self.environment

        # Add each thread-local attribute to record.
        for attr in dir(self.local):
            if attr.startswith("__") and attr.endswith("__"):
                continue
            setattr(record, attr, getattr(self.local, attr))
        if not hasattr(record, "identifier"):
            record.identifier = "Unknown"

        return True



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library debug options.

DEPRECATION WARNING: Some parameters which don't belong here will be removed.
"""
import os
from pathlib import Path
from collections import deque


class Debug:
    """Debug flags for ETOS."""

    __events_published = deque(
        maxlen=int(os.getenv("ETOS_PUBLISHED_EVENT_HISTORY_SIZE", "100"))
    )
    __events_received = deque(
        maxlen=int(os.getenv("ETOS_RECEIVED_EVENT_HISTORY_SIZE", "100"))
    )

    @property
    def default_secret_path(self):
        """Path to k8s secrets."""
        return os.getenv("ETOS_SECRET_PATH", "/etc/")

    @property
    def default_log_path(self):
        """"Default log path."""
        path = os.getenv("ETOS_LOG_PATH")
        if path is None:
            path = Path.home().joinpath("logs/log.json")
        else:
            path = Path(path)
        return path

    @property
    def disable_sending_events(self):
        """Disable sending eiffel events."""
        return bool(os.getenv("ETOS_DISABLE_SENDING_EVENTS", None))

    @property
    def disable_receiving_events(self):
        """Disable receiving eiffel events."""
        return bool(os.getenv("ETOS_DISABLE_RECEIVING_EVENTS", None))

    @property
    def default_http_timeout(self):
        """Timeout for HTTP requests."""
        # Defaults to 1h
        return int(os.getenv("ETOS_DEFAULT_HTTP_TIMEOUT", "3600"))

    @property
    def default_wait_timeout(self):
        """Timeout for utils.wait."""
        # Defaults to 60s
        return int(os.getenv("ETOS_DEFAULT_WAIT_TIMEOUT", "60"))

    @property
    def default_test_result_timeout(self):
        """Timeout for waiting for test results."""
        # Defaults to 24h
        return int(os.getenv("ETOS_DEFAULT_TEST_RESULT_TIMEOUT", "86400"))

    @property
    def debug_published_event_history_size(self):
        """Maximum number of published events to store in history."""
        return int(os.getenv("ETOS_PUBLISHED_EVENT_HISTORY_SIZE", "100"))

    @property
    def debug_received_event_history_size(self):
        """Maximum number of published events to store in history."""
        return int(os.getenv("ETOS_RECEIVED_EVENT_HISTORY_SIZE", "100"))

    @property
    def debug_store_events_published(self):
        """Store all eiffel events published. Do not use in production."""
        print("debug_store_events_published is DEPRECATED")
        return bool(os.getenv("ETOS_DEBUG_STORE_EVENTS_PUBLISHED", None))

    @property
    def debug_store_events_received(self):
        """Store all eiffel events received. Do not use in production."""
        print("debug_store_events_received is DEPRECATED")
        return bool(os.getenv("ETOS_DEBUG_STORE_EVENTS_RECEIVED", None))

    @property
    def disable_monitor_announce(self):
        """Disable the announcement event from monitor."""
        return bool(os.getenv("ETOS_DISABLE_MONITOR_ANNOUNCE", None))

    @property
    def environment_provider(self):
        """Environment provider host to use."""
        if os.getenv("ETOS_ENVIRONMENT_PROVIDER") is not None:
            return os.getenv("ETOS_ENVIRONMENT_PROVIDER")
        raise Exception("ETOS_ENVIRONMENT_PROVIDER environment variable not set!")

    @property
    def etos_api(self):
        """ETOS API host to use."""
        if os.getenv("ETOS_API") is not None:
            return os.getenv("ETOS_API")
        raise Exception("ETOS_API environment variable not set!")

    @property
    def graphql_server(self):
        """Graphql server to use."""
        if os.getenv("ETOS_GRAPHQL_SERVER") is not None:
            return os.getenv("ETOS_GRAPHQL_SERVER")
        raise Exception("ETOS_GRAPHQL_SERVER environment variable not set!")

    @property
    def database_host(self):
        """Database host."""
        return os.getenv("ETOS_DATABASE_HOST", "localhost")

    @property
    def database_port(self):
        """Database port."""
        return int(os.getenv("ETOS_DATABASE_PORT", "26379"))

    @property
    def events_published(self):
        """All events published if ETOS_DEBUG_STORE_EVENTS_PUBLISHED flag is set."""
        return self.__events_published

    @property
    def events_received(self):
        """All events received if ETOS_DEBUG_STORE_EVENTS_RECEIVED flag is set."""
        return self.__events_received



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Library HTTP Client."""
from json.decoder import JSONDecodeError
import logging
import time
import traceback
from urllib3.exceptions import MaxRetryError, NewConnectionError
import requests
from requests.exceptions import HTTPError
from .debug import Debug


class Http:
    """Utility class for HTTP requests."""

    logger = logging.getLogger("HTTP")

    def __init__(self):
        """Initialize a debug instance."""
        self.debug = Debug()

    @staticmethod
    def request(verb, url, as_json=True, **requests_kwargs):
        """Make an HTTP request.

        :param verb: Which HTTP verb to use. GET, PUT, POST
                     (DELETE omitted)
        :type verb: str
        :param as_json: Whether or not to return json instead of response.
        :type as_json: bool
        :param request_kwargs: Keyword arguments for the requests command.
        :type request_kwargs: dict
        :return: HTTP response or json.
        :rtype: Response or dict
        """
        request = None
        if verb.lower() == "put":
            request = requests.put
        elif verb.lower() == "post":
            request = requests.post
        elif verb.lower() == "get":
            request = requests.get
        response = request(url, **requests_kwargs)
        response.raise_for_status()
        if as_json:
            return response.json()
        return response

    def retry(self, verb, url, timeout=None, as_json=True, **requests_kwargs):
        """Attempt to connect to url for x time.

        :param verb: Which HTTP verb to use. GET, PUT, POST
                     (DELETE omitted)
        :type verb: str
        :param timeout: How long, in seconds, to retry request.
        :type timeout: int or None
        :param as_json: Whether or not to return json instead of response.
        :type as_json: bool
        :param request_kwargs: Keyword arguments for the requests command.
        :type request_kwargs: dict
        :return: HTTP response or json.
        :rtype: Response or dict
        """
        if timeout is None:
            timeout = self.debug.default_http_timeout
        end_time = time.time() + timeout
        self.logger.debug(
            "Retrying URL %s for %d seconds with a %s request.", url, timeout, verb
        )
        iteration = 0
        while time.time() < end_time:
            iteration += 1
            self.logger.debug("Iteration: %d", iteration)
            try:
                yield self.request(verb, url, as_json, **requests_kwargs)
                break
            except (
                ConnectionError,
                HTTPError,
                NewConnectionError,
                MaxRetryError,
                TimeoutError,
                JSONDecodeError,
            ):
                traceback.print_exc()
                time.sleep(2)
        else:
            raise ConnectionError(
                "Unable to {} {} with params {}".format(verb, url, requests_kwargs)
            )

    def wait_for_request(self, uri, timeout=None, as_json=True, **params):
        """Request a URI with parameters until it succeeds.

        :param uri: URI to request.
        :type uri: str
        :param timeout: How long, in seconds, to attempt request.
        :type timeout: int
        :param as_json: Whether the response is json or not.
        :type as_json: bool
        :param params: Parameters for the request.
        :type params: dict
        :return: HTTP response or json.
        :rtype: Response or dict
        """
        self.logger.info("Requesting '%s' with parameters '%s'", uri, params)
        return self.retry("GET", uri, timeout, as_json, **params)



# Copyright 2020 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS logger.

Example::

    import logging
    from uuid import uuid4
    from etos_lib.logging.logger import setup_logging, FORMAT_CONFIG

    FORMAT_CONFIG.identifier = str(uuid4())
    setup_logging("myApp", "1.0.0", "production")
    logger = logging.getLogger(__name__)
    logger.info("Hello!")
    >>> [2020-12-16 10:35:00][cb7c8cd9-40a6-4ecc-8321-a1eae6beae35] INFO: Hello!

"""
import sys
from pathlib import Path
import threading
import logging
import logging.config
from box import Box
from etos_lib.logging.filter import EtosFilter
from etos_lib.logging.formatter import EtosLogFormatter
from etos_lib.lib.debug import Debug

DEFAULT_CONFIG = Path(__file__).parent.joinpath("default_config.yaml")
DEFAULT_LOG_PATH = Debug().default_log_path

FORMAT_CONFIG = threading.local()


def setup_file_logging(config, log_filter):
    """Set up logging to file using the ETOS log formatter.

    Cofiguration file parameters ('file' must exist or no file handler is set up):

        logging:
          file:
            # Log level for file logging. Default=DEBUG.
            loglevel: INFO
            # Where to store logfile. Default=/home/you/etos/output.log.json
            logfile: path/to/log/file
            # Maximum number of files to rotate. Default=10
            max_files: 5
            # Maximum number of bytes in each logfile. Default=1048576/1MB
            max_bytes: 100

    :param config: File logging configuration.
    :type config: :obj:`Box`
    :param log_filter: Logfilter to add to file handler.
    :type log_filter: :obj:`EtosFilter`
    """
    loglevel = getattr(logging, config.get("loglevel", "DEBUG"))
    logfile = Path(config.get("logfile", DEFAULT_LOG_PATH))
    logfile.parent.mkdir(parents=True, exist_ok=True)

    max_files = config.get("max_files", 10)
    max_bytes = config.get("max_bytes", 10485760)  # Default is 10 MB
    root_logger = logging.getLogger()

    file_handler = logging.handlers.RotatingFileHandler(
        logfile, maxBytes=max_bytes, backupCount=max_files
    )
    file_handler.setFormatter(EtosLogFormatter())
    file_handler.setLevel(loglevel)
    file_handler.addFilter(log_filter)
    root_logger.addHandler(file_handler)


def setup_stream_logging(config, log_filter):
    """Set up logging to stdout stream.

    Cofiguration file parameters ('stream' must exist or no stream handler is set up):

        logging:
          stream:
            # Log level for stream logging. Default=INFO.
            loglevel: ERROR
            # Format to print logs with.
            # Default: [%(asctime)s][%(identifier)s] %(levelname)s:%(name)s: %(message)s
            logformat: %(message)s
            # Dateformat for %(asctime) format. Default: %Y-%m-%d %H:%M:%S
            dateformat: %Y-%d-%m %H:%M:%S

    :param config: Stream logging configuration.
    :type config: :obj:`Box`
    :param log_filter: Logfilter to add to stream handler.
    :type log_filter: :obj:`EtosFilter`
    """
    loglevel = getattr(logging, config.get("loglevel", "INFO"))

    logformat = config.get(
        "logformat",
        "[%(asctime)s][%(identifier)s] %(levelname)s:%(name)s: %(message)s"
    )
    dateformat = config.get("dateformat", "%Y-%m-%d %H:%M:%S")
    root_logger = logging.getLogger()
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(logging.Formatter(logformat, datefmt=dateformat))
    stream_handler.setLevel(loglevel)
    stream_handler.addFilter(log_filter)
    root_logger.addHandler(stream_handler)


def setup_logging(
    application, version, environment, config_file=DEFAULT_CONFIG
):
    """Set up basic logging.

    :param application: Name of application to setup logging for.
    :type application: str
    :param version: Version of application to setup logging for.
    :type version: str
    :param environment: Environment in which this application resides.
    :type environment: str
    :param config_file: Filename of logging configuration.
    :type config_file: str
    """
    with open(config_file) as yaml_file:
        config = Box.from_yaml(yaml_file)
    logging_config = config.logging

    log_filter = EtosFilter(application, version, environment, FORMAT_CONFIG)

    # Create a default logger which will not propagate messages
    # to the root logger. This logger will create records for all
    # messages, but not print them to stdout. Stdout printing
    # is setup in "setup_stream_logging" if the "stream" key exists
    # in the configuration file.
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)
    root_logger.propagate = 0

    if logging_config.get("stream"):
        setup_stream_logging(logging_config.get("stream"), log_filter)
    if logging_config.get("file"):
        setup_file_logging(logging_config.get("file"), log_filter)
